/****************************************************************************
 *
 * KanjiHack v1.0
 * by Dan Potter (bard@mail.utexas.edu)
 *
 * Copyright 1998 Dan Potter. No restrictions are placed on this code,
 * as long as the copyright notice is preserved, and you have fun using it ;)
 *
 ****************************************************************************
 * 
 * This module is a "text drawing daemon". All requests to
 *  draw text go through this module, so that EUC text may be detected and
 *  drawn correctly. This is a work-around hack for those people who either
 *  don't have the patience, or don't have the capability to get proper
 *  locale support for Japanese. It may work with other double-byte languages
 *  after a little hacking. It has a few caveats though. For one, text width
 *  is calculated incorrectly in strings that have kanji. For another, there
 *  is only one kanji font possible (although, being a native English speaker,
 *  I find this to be far less limiting than only one roman font =). 
 *
 * !!!seriously disclaimer!!! No, my grammar didn't just falter.. we got a 
 *  funny 100bT hub at work that said "seriously warning!" on the bottom. 
 *  Anyway =), I am a native English speaker, and I've only been studying
 *  Japanese for about half a year now off and on. Also, I have only very
 *  recently gotten into Japanese encoding on computers. So there are probably
 *  some problems with this! I use this mainly as an educational tool, so 
 *  people who really use it may find things that don't work (such as, this
 *  in theory only supports EUC Code Set 1, the Jouyou kanji).
 *
 * This version has been adapted from its original AfterStep home to work
 *  window WindowMaker. In order to keep from screwing too many things up 
 *  in this version (and to make it easier for people to put this patch in
 *  later versions), the EUC font is currently HARDCODED. You should set these
 *  two options in wconfig.h if you wish to have support for this module
 *  compiled in:
 *
 *  #define XDrawString KANJIHACK_DrawString
 *  #define KANJIHACK_FONT  "k14"
 *
 * The second line is variable, the first is not, unless you rename the 
 *  function below. You should also disable any compiled-in I18N support, as
 *  this will probably hamper XDrawString from getting its characters. If
 *  you leave out the top #define, the module may get compiled in but won't
 *  have any effect.
 *
 ***************************************************************************/

/* I'm not sure how many of these headers we really need, but it won't really
 * hurt to include too many.. This list is taken from configure.c. */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xproto.h>
#include <X11/Xatom.h>

#include "wconfig.h"

/*
  XDrawString16(Display *, Drawable, GC, x, y, XChar2b *string, int len);

Code set 0 (ASCII):
  first byte	0x21-0x7E
Code set 1 (JIS X 0208-1990):
  first byte	0xA1-0xFE
  second byte	0xA1-0xFE
Code set 2 (Half-width katakana):
  first byte	0x8E
  second byte	0xA1-0xDF
Code set 3 (User-defined characters):
  first byte	0x8F
  second byte	0xA1-0xFE

 */

// This expression defines whether a character can be the starting of an
// EUC two-byte sequence.
//#define isEUC(A) ((0xA1<=(A)<=0xFE || (A)==0x8E || (A)==0x8F) ? 1 : 0)

// This one only detects code set 1 right now (which is all we support right 
// now). Anyone want to add other support (half-width katakana, etc)?
#define isEUC(A) ((0xA1<=(A)) && ((A)<=0xFE) ? 1 : 0)

// This represents our EUC font. Again, a hack, but this is trying to keep
// the actual WindowMaker sources clean.
static XFontStruct *eucfontstr = NULL;


// Returns the EUC-decoded value of a two-byte EUC sequence.
// Currently only supports code set 1 (I think).
static void decodeEUC(int *b1, int *b2) {
	*b1 -= 0x80;
	*b2 -= 0x80;
}


// This function is meant as a drop-in replacement for XDrawString. However,
// before drawing any text, it scans for valid EUC codes, and draws each
// segment of the string that uses EUC with XDrawString16 and the specified
// EUC font (if there was one!). Things will poop out if it can't load the
// EUC font. In this case, it just calls the old XDrawString.

// The powers that be help us if we forget this little hack.. =)
#undef XDrawString

int KANJIHACK_DrawString(Display *dpy, Drawable d, GC gc, int x, int y, 
		_Xconst char *str, int len) {
	int	q, eucyes;
	char	ourstr[256];
	char	tmpstr[256];
	XChar2b tmpstr2[256];
	int	tmpstrlen;
	Font	ascfont, eucfont;
	XFontStruct *ascfontstr;
	XGCValues gcv;

  #define ustr ((unsigned char *)str)

  // No text? Screw it..
	if (!len) return 0;

  // First things first. If there is no EUC text involved, then don't even
  // bother to do anything funky.
	eucyes = 0;
	for (q=0; q<len; q++) {
		if (isEUC(ustr[q])) {
			eucyes = 1;
			break;	// we've got EUC
		}
	}

  // Did we load the EUC font yet?
	if (eucfontstr == (XFontStruct*)-1) {
		eucyes = 0;
	} else if (!eucfontstr) {
		eucfontstr = XLoadQueryFont(dpy, KANJIHACK_FONT);
		if (!eucfontstr) {
		  // I swear.. it's not my grammar.. it's a bad joke =)
			fprintf(stderr, " !!!seriously warning!!! Can't load euc font %s\n", KANJIHACK_FONT);
			eucfontstr = (XFontStruct*)-1;
			eucyes = 0;
		}
	}

  // If there was no EUC text, then call XDrawString directly and quit.
	if (!eucyes) {
		return XDrawString(dpy, d, gc, x, y, str, len);
	}

  // Make a copy of the input string for us (which is writable)
	strncpy(ourstr, str, len);
	str = ourstr; str[len]='\0';	// ignore the warning!

  // Set up font structs for easy access
	XGetGCValues(dpy, gc, GCFont, &gcv);
	ascfont = gcv.font;
	eucfont = eucfontstr->fid;

	ascfontstr = XQueryFont(dpy, ascfont);

  // We'll chop the string up into parts and draw them appropriately.
	while (*str) {
	   // Is this segment EUC or ASCII?
		if (isEUC(*ustr)) {		// EUC
		   // Dump that into the tmpstr buffer until we've got it all.
		   // When it's all there, print it and scoot the pointers 
		   // forward again.
			q = 0;
			while (*str) {
			   // 16-bit fonts are organized as a matrix
				int row = *str++;
				int col = *str++;

				decodeEUC(&row, &col);

				if (col == 0) {
				   // Chopped off two-byte sequence: quit.
					str--;
					break;
				}

				tmpstr2[q].byte1 = (unsigned char)row;
				tmpstr2[q].byte2 = (unsigned char)col;
				q++;

			   // If the next char is not an EUC sequence, then jump
			   // back to normal ASCII mode.
				if (!isEUC(*ustr)) {
				   // We've hit the end of the EUC. So print it.
					tmpstrlen = q;
					XSetFont(dpy, gc, eucfont);
					XDrawString16(dpy, d, gc, x, y, 
						tmpstr2, tmpstrlen);
					x += XTextWidth16(eucfontstr, 
						tmpstr2, tmpstrlen);
					XSetFont(dpy, gc, ascfont);
					break;
				}
			}
		}
		else {				// ASCII
		   // Copy the ASCII part of the string and null-terminate it.
			q = 0;
			while (*str) {
				if (!isEUC(*ustr)) {
					tmpstr[q++] = *str;
					str++;
				}
				else
					break;
			}
			tmpstr[q] = '\0';
			tmpstrlen = q;

		   // Draw that part using the user-selected font for the
		   // particular element.
			XDrawString(dpy, d, gc, x, y, tmpstr, tmpstrlen);

		   // Move the X output position
			x += XTextWidth(ascfontstr, tmpstr, tmpstrlen);
		}

	}

   // EUC font operations are bracketed by SetFont commands, so we should
   // be safely back in ascii font mode now.
	XFreeFontInfo(NULL,ascfontstr,0);	// we alloc'd this one

	return 0;
}
